from __future__ import annotations

import itertools
import math
import random
from typing import Any

import pandas as pd
import pytest

from vivarium.framework.population import PopulationError, PopulationManager, PopulationView

##########################
# Mock data and fixtures #
##########################

COL_NAMES = ["color", "count", "pie", "pi", "tracked"]
COLORS = ["red", "green", "yellow"]
COUNTS = [10, 20, 30]
PIES = ["apple", "chocolate", "pecan"]
PIS = [math.pi**i for i in range(1, 4)]
TRACKED_STATUSES = [True, False]
RECORDS = [
    (color, count, pie, pi, ts)
    for color, count, pie, pi, ts in itertools.product(
        COLORS, COUNTS, PIES, PIS, TRACKED_STATUSES
    )
]
BASE_POPULATION = pd.DataFrame(data=RECORDS, columns=COL_NAMES)

NEW_COL_NAMES = ["cube", "cube_string"]
CUBE = [i**3 for i in range(len(RECORDS))]
CUBE_STRING = [str(i**3) for i in range(len(RECORDS))]
NEW_ATTRIBUTES = pd.DataFrame(
    zip(CUBE, CUBE_STRING),
    columns=NEW_COL_NAMES,
    index=BASE_POPULATION.index,
)


@pytest.fixture(scope="function")
def population_manager() -> PopulationManager:
    class _PopulationManager(PopulationManager):
        def __init__(self) -> None:
            super().__init__()
            self._population = pd.DataFrame(
                data=RECORDS,
                columns=COL_NAMES,
            )

        def _add_constraint(self, *args: Any, **kwargs: Any) -> None:
            pass

    return _PopulationManager()


@pytest.fixture(
    params=[
        BASE_POPULATION.index,
        BASE_POPULATION.index[::2],
        BASE_POPULATION.index[:0],
    ]
)
def update_index(request: pytest.FixtureRequest) -> pd.Index[int]:
    index = request.param
    assert isinstance(index, pd.Index)
    return index


@pytest.fixture(
    params=[
        BASE_POPULATION.copy(),
        BASE_POPULATION[COL_NAMES[:2]].copy(),
        BASE_POPULATION[[COL_NAMES[0]]].copy(),
        BASE_POPULATION[COL_NAMES[0]].copy(),
    ]
)
def population_update(
    request: pytest.FixtureRequest, update_index: pd.Index[int]
) -> pd.Series[Any] | pd.DataFrame:
    update = request.param.loc[update_index]
    assert isinstance(update, (pd.Series, pd.DataFrame))
    return update


@pytest.fixture(
    params=[
        NEW_ATTRIBUTES.copy(),
        NEW_ATTRIBUTES[[NEW_COL_NAMES[0]]].copy(),
        NEW_ATTRIBUTES[NEW_COL_NAMES[0]].copy(),
        pd.concat([BASE_POPULATION, NEW_ATTRIBUTES], axis=1),
        pd.concat([BASE_POPULATION.iloc[:, 0], NEW_ATTRIBUTES.iloc[:, 0]], axis=1),
    ]
)
def population_update_new_cols(
    request: pytest.FixtureRequest, update_index: pd.Index[int]
) -> pd.Series[Any] | pd.DataFrame:
    update = request.param.loc[update_index]
    assert isinstance(update, (pd.Series, pd.DataFrame))
    return update


##################
# Initialization #
##################


def test_initialization(population_manager: PopulationManager) -> None:
    pv = population_manager.get_view(COL_NAMES)
    assert pv._id == 0
    assert pv.name == "population_view_0"
    assert set(pv.columns) == set(COL_NAMES)
    assert pv.query == ""

    # Failure here is lazy.  The manager should give you back views for
    # columns that don't exist since views are built during setup when
    # we don't necessarily know all the columns yet.
    cols = ["age", "sex", "tracked"]
    pv = population_manager.get_view(cols)
    assert pv._id == 1
    assert pv.name == "population_view_1"
    assert set(pv.columns) == set(cols)
    assert pv.query == ""

    col_subset = ["color", "count"]
    pv = population_manager.get_view(col_subset)
    assert pv._id == 2
    assert pv.name == "population_view_2"
    assert set(pv.columns) == set(col_subset)
    # View will filter to tracked by default if it's not requested as a column
    assert pv.query == "tracked == True"

    q_string = "color == 'red'"
    pv = population_manager.get_view(COL_NAMES, query=q_string)
    assert pv._id == 3
    assert pv.name == "population_view_3"
    assert set(pv.columns) == set(COL_NAMES)
    assert pv.query == q_string


##########################
# PopulationView.subview #
##########################


@pytest.mark.parametrize(
    "columns",
    [["color", "count"], ["color"], ["color", "count", "tracked"]],
    ids=["multiple columns", "single column", "including tracked"],
)
def test_subview_columns_list_input(
    population_manager: PopulationManager, columns: list[str]
) -> None:
    pv = population_manager.get_view(COL_NAMES)
    sub_pv = pv.subview(columns)
    assert set(sub_pv.columns) == set(columns)


def test_subview_columns_string_input(population_manager: PopulationManager) -> None:
    pv = population_manager.get_view(COL_NAMES)
    sub_pv = pv.subview("color")
    assert set(sub_pv.columns) == {"color"}


@pytest.mark.parametrize("columns", [["color", "count"], ["color", "count", "tracked"]])
def test_subview_queries(population_manager: PopulationManager, columns: list[str]) -> None:
    pv = population_manager.get_view(COL_NAMES, query="foo == 'red'")

    # get subview
    sub_pv = pv.subview(columns)

    expected_query = pv.query if "tracked" in columns else f"{pv.query} and tracked == True"
    assert sub_pv.query == expected_query


@pytest.mark.parametrize(
    "columns",
    [["age", "sex"], COL_NAMES + ["age"], []],
    ids=["columns not in pop_view", "one column not in pop_view", "no columns"],
)
def test_subview_bad_columns_input(
    population_manager: PopulationManager, columns: list[str]
) -> None:
    pv = population_manager.get_view(COL_NAMES)
    with pytest.raises(PopulationError):
        pv.subview(columns)


######################
# PopulationView.get #
######################


def test_get(population_manager: PopulationManager) -> None:
    ########################
    # Full population view #
    ########################
    pv = population_manager.get_view(COL_NAMES)
    full_idx = pd.RangeIndex(0, len(RECORDS))

    # Get full data set
    pop = pv.get(full_idx)
    assert set(pop.columns) == set(COL_NAMES)
    assert len(pop) == len(RECORDS)

    # Get data subset
    pop = pv.get(full_idx, query=f"color == 'red'")
    assert set(pop.columns) == set(COL_NAMES)
    assert len(pop) == len(RECORDS) // len(COLORS)

    ###############################
    # View without tracked column #
    ###############################
    cols_without_tracked = COL_NAMES[:-1]
    pv = population_manager.get_view(cols_without_tracked)

    # Get all tracked
    pop = pv.get(full_idx)
    assert set(pop.columns) == set(cols_without_tracked)
    assert len(pop) == len(RECORDS) // 2

    # get subset without tracked
    pop = pv.get(full_idx, query=f"color == 'red'")
    assert set(pop.columns) == set(cols_without_tracked)
    assert len(pop) == len(RECORDS) // (2 * len(COLORS))


def test_get_empty_idx(population_manager: PopulationManager) -> None:
    pv = population_manager.get_view(COL_NAMES)

    pop = pv.get(pd.Index([]))
    assert isinstance(pop, pd.DataFrame)
    assert set(pop.columns) == set(COL_NAMES)
    assert pop.empty


def test_get_fail(population_manager: PopulationManager) -> None:
    bad_pvs = [
        population_manager.get_view(["age", "sex"]),
        population_manager.get_view(COL_NAMES + ["age", "sex"]),
        population_manager.get_view(["age", "sex", "tracked"]),
        population_manager.get_view(["age", "sex"]),
        population_manager.get_view(["color", "count", "age"]),
    ]

    full_idx = pd.RangeIndex(0, len(RECORDS))

    for pv in bad_pvs:
        with pytest.raises(PopulationError, match="not in population table"):
            pv.get(full_idx)


#################################
# PopulationView.update helpers #
#######################################
# PopulationView._coerce_to_dataframe #
#######################################


def test_full_population_view__coerce_to_dataframe(
    population_update: pd.Series[Any] | pd.DataFrame,
    update_index: pd.Index[int],
) -> None:
    if isinstance(population_update, pd.Series):
        cols = [population_update.name]
    else:
        cols = list(population_update.columns)
    coerced_df = PopulationView._coerce_to_dataframe(population_update, COL_NAMES)
    assert BASE_POPULATION.loc[update_index, cols].equals(coerced_df)


def test_full_population_view__coerce_to_dataframe_fail(
    population_update_new_cols: pd.Series[Any] | pd.DataFrame,
) -> None:
    with pytest.raises(TypeError, match="must be a pandas Series or DataFrame"):
        PopulationView._coerce_to_dataframe(BASE_POPULATION.iloc[:, 0].tolist(), COL_NAMES)  # type: ignore[arg-type]

    with pytest.raises(PopulationError, match="unnamed pandas series"):
        PopulationView._coerce_to_dataframe(
            BASE_POPULATION.iloc[:, 0].rename(None),
            COL_NAMES,
        )

    with pytest.raises(PopulationError, match="extra columns"):
        PopulationView._coerce_to_dataframe(population_update_new_cols, COL_NAMES)

    with pytest.raises(PopulationError, match="no columns"):
        PopulationView._coerce_to_dataframe(
            pd.DataFrame(index=BASE_POPULATION.index), COL_NAMES
        )


def test_single_column_population_view__coerce_to_dataframe(
    update_index: pd.Index[int],
) -> None:
    column = COL_NAMES[0]
    update = BASE_POPULATION.loc[update_index].copy()
    output = BASE_POPULATION.loc[update_index, [column]]

    passing_cases = [
        update[[column]],  # Single col df
        update[column],  # named series
        update[column].rename(None),  # Unnamed series
    ]
    for case in passing_cases:
        assert isinstance(case, (pd.Series, pd.DataFrame))
        coerced_df = PopulationView._coerce_to_dataframe(case, [column])
        assert output.equals(coerced_df)


def test_single_column_population_view__coerce_to_dataframe_fail(
    population_update_new_cols: pd.Series[Any] | pd.DataFrame,
) -> None:
    with pytest.raises(TypeError, match="must be a pandas Series or DataFrame"):
        PopulationView._coerce_to_dataframe(
            BASE_POPULATION.iloc[:, 0].tolist(), [COL_NAMES[0]]  # type: ignore[arg-type]
        )

    with pytest.raises(PopulationError, match="extra columns"):
        PopulationView._coerce_to_dataframe(population_update_new_cols, [COL_NAMES[0]])

    with pytest.raises(PopulationError, match="no columns"):
        PopulationView._coerce_to_dataframe(
            pd.DataFrame(index=BASE_POPULATION.index), [COL_NAMES[0]]
        )


#################################
# PopulationView.update helpers #
##################################################
# PopulationView._ensure_coherent_initialization #
##################################################


def test__ensure_coherent_initialization_no_new_columns(
    population_update: pd.Series[Any] | pd.DataFrame,
    update_index: pd.Index[int],
) -> None:
    if isinstance(population_update, pd.Series):
        pytest.skip()

    # Missing population
    if not update_index.empty:
        with pytest.raises(PopulationError, match="missing updates"):
            PopulationView._ensure_coherent_initialization(
                population_update.loc[update_index[::2]], BASE_POPULATION.loc[update_index]
            )

    # No new columns
    with pytest.raises(PopulationError, match="all provided columns"):
        PopulationView._ensure_coherent_initialization(
            population_update,
            BASE_POPULATION.loc[update_index],
        )


def test__ensure_coherent_initialization_new_columns(
    population_update_new_cols: pd.Series[Any] | pd.DataFrame,
    update_index: pd.Index[int],
) -> None:
    if isinstance(population_update_new_cols, pd.Series):
        pytest.skip()

    # All new cols, should be good
    PopulationView._ensure_coherent_initialization(
        population_update_new_cols,
        BASE_POPULATION.loc[update_index],
    )

    # Missing rows
    if not update_index.equals(BASE_POPULATION.index):
        with pytest.raises(PopulationError, match="missing updates"):
            PopulationView._ensure_coherent_initialization(
                population_update_new_cols,
                BASE_POPULATION,
            )

    # Conflicting data in existing cols.
    cols_overlap = [c for c in population_update_new_cols if c in COL_NAMES]
    if not update_index.empty and cols_overlap:
        update = population_update_new_cols.copy()
        update[COL_NAMES[0]] = "bad_values"
        with pytest.raises(PopulationError, match="conflicting"):
            PopulationView._ensure_coherent_initialization(
                update,
                BASE_POPULATION.loc[update_index],
            )


#################################
# PopulationView.update helpers #
#########################################################
# PopulationView._format_update_and_check_preconditions #
#########################################################


def test__format_update_and_check_preconditions_bad_args() -> None:
    with pytest.raises(AssertionError):
        PopulationView._format_update_and_check_preconditions(
            BASE_POPULATION,
            BASE_POPULATION,
            COL_NAMES,
            creating_initial_population=True,
            adding_simulants=False,
        )

    with pytest.raises(TypeError, match="must be a pandas Series or DataFrame"):
        PopulationView._format_update_and_check_preconditions(
            BASE_POPULATION.iloc[:, 0].tolist(),  # type: ignore[arg-type]
            BASE_POPULATION,
            COL_NAMES,
            True,
            True,
        )


def test__format_update_and_check_preconditions_coerce_failures(
    population_update_new_cols: pd.Series[Any] | pd.DataFrame,
) -> None:
    with pytest.raises(PopulationError, match="unnamed pandas series"):
        PopulationView._format_update_and_check_preconditions(
            BASE_POPULATION.iloc[:, 0].rename(None),
            BASE_POPULATION,
            COL_NAMES,
            True,
            True,
        )

    for view_cols in [COL_NAMES, [COL_NAMES[0]]]:
        with pytest.raises(PopulationError, match="extra columns"):
            PopulationView._format_update_and_check_preconditions(
                population_update_new_cols,
                BASE_POPULATION,
                view_cols,
                True,
                True,
            )

        with pytest.raises(PopulationError, match="no columns"):
            PopulationView._format_update_and_check_preconditions(
                pd.DataFrame(index=BASE_POPULATION.index),
                BASE_POPULATION,
                view_cols,
                True,
                True,
            )


def test__format_update_and_check_preconditions_unknown_pop_fail(
    population_update: pd.Series[Any] | pd.DataFrame,
) -> None:
    if population_update.empty:
        pytest.skip()

    update = population_update.copy()
    update.index += 2 * update.index.max()

    with pytest.raises(PopulationError, match=f"{len(update)} simulants"):
        PopulationView._format_update_and_check_preconditions(
            update,
            BASE_POPULATION,
            COL_NAMES,
            True,
            True,
        )


def test__format_update_and_check_preconditions_coherent_initialization_fail(
    population_update: pd.Series[Any] | pd.DataFrame,
    update_index: pd.Index[int],
) -> None:
    # Missing population
    if not update_index.empty:
        with pytest.raises(PopulationError, match="missing updates"):
            PopulationView._format_update_and_check_preconditions(
                population_update.loc[update_index[::2]],
                BASE_POPULATION.loc[update_index],
                COL_NAMES,
                True,
                True,
            )

    # No new columns
    with pytest.raises(PopulationError, match="all provided columns"):
        PopulationView._format_update_and_check_preconditions(
            population_update,
            BASE_POPULATION.loc[update_index],
            COL_NAMES,
            True,
            True,
        )


def test__format_update_and_check_preconditions_coherent_initialization_fail_new_cols(
    population_update_new_cols: pd.Series[Any] | pd.DataFrame,
    update_index: pd.Index[int],
) -> None:
    if not update_index.equals(BASE_POPULATION.index):
        with pytest.raises(PopulationError, match="missing updates"):
            PopulationView._format_update_and_check_preconditions(
                population_update_new_cols,
                BASE_POPULATION,
                COL_NAMES + NEW_COL_NAMES,
                True,
                True,
            )

    # Conflicting data in existing cols.
    cols_overlap = [c for c in population_update_new_cols if c in COL_NAMES]
    if not update_index.empty and cols_overlap:
        update = population_update_new_cols.copy()
        update[COL_NAMES[0]] = "bad_values"
        with pytest.raises(PopulationError, match="conflicting"):
            PopulationView._format_update_and_check_preconditions(
                update,
                BASE_POPULATION.loc[update_index],
                COL_NAMES + NEW_COL_NAMES,
                True,
                True,
            )


def test__format_update_and_check_preconditions_new_columns_non_init(
    population_update_new_cols: pd.Series[Any] | pd.DataFrame,
    update_index: pd.Index[int],
) -> None:
    for adding_simulants in [True, False]:
        with pytest.raises(PopulationError, match="outside the initial population creation"):
            PopulationView._format_update_and_check_preconditions(
                population_update_new_cols,
                BASE_POPULATION.loc[update_index],
                COL_NAMES + NEW_COL_NAMES,
                False,
                adding_simulants,
            )


def test__format_update_and_check_preconditions_conflicting_non_init(
    population_update: pd.Series[Any] | pd.DataFrame,
    update_index: pd.Index[int],
) -> None:
    update = population_update.copy()
    if isinstance(update, pd.Series):
        update[:] = "bad_value"
    else:
        update.loc[:, COL_NAMES[0]] = "bad_value"
    if not update_index.empty:
        with pytest.raises(PopulationError, match="conflicting"):
            PopulationView._format_update_and_check_preconditions(
                update,
                BASE_POPULATION.loc[update_index],
                COL_NAMES + NEW_COL_NAMES,
                False,
                True,
            )


def test__format_update_and_check_preconditions_init_pass(
    population_update_new_cols: pd.Series[Any] | pd.DataFrame,
    update_index: pd.Index[int],
) -> None:
    result = PopulationView._format_update_and_check_preconditions(
        population_update_new_cols,
        BASE_POPULATION.loc[update_index],
        COL_NAMES + NEW_COL_NAMES,
        True,
        True,
    )
    update = PopulationView._coerce_to_dataframe(
        population_update_new_cols,
        COL_NAMES + NEW_COL_NAMES,
    )

    assert set(result.columns) == set(update)
    for col in update:
        assert result[col].equals(update[col])


def test__format_update_and_check_preconditions_add_pass(
    population_update: pd.Series[Any] | pd.DataFrame,
    update_index: pd.Index[int],
) -> None:
    state_table = BASE_POPULATION.drop(update_index).reindex(BASE_POPULATION.index)
    result = PopulationView._format_update_and_check_preconditions(
        population_update,
        state_table,
        COL_NAMES + NEW_COL_NAMES,
        False,
        True,
    )
    update = PopulationView._coerce_to_dataframe(
        population_update,
        COL_NAMES + NEW_COL_NAMES,
    )

    assert set(result.columns) == set(update)
    for col in update:
        assert result[col].equals(update[col])


def test__format_update_and_check_preconditions_time_step_pass(
    population_update: pd.Series[Any] | pd.DataFrame,
) -> None:
    result = PopulationView._format_update_and_check_preconditions(
        population_update,
        BASE_POPULATION,
        COL_NAMES + NEW_COL_NAMES,
        False,
        False,
    )
    update = PopulationView._coerce_to_dataframe(
        population_update,
        COL_NAMES + NEW_COL_NAMES,
    )

    assert set(result.columns) == set(update)
    for col in update:
        assert result[col].equals(update[col])


def test__format_update_and_check_preconditions_adding_simulants_replace_identical_data(
    population_update: pd.Series[Any] | pd.DataFrame,
) -> None:
    result = PopulationView._format_update_and_check_preconditions(
        population_update,
        BASE_POPULATION,
        COL_NAMES + NEW_COL_NAMES,
        False,
        True,
    )
    update = PopulationView._coerce_to_dataframe(
        population_update,
        COL_NAMES + NEW_COL_NAMES,
    )

    assert set(result.columns) == set(update)
    for col in update:
        assert result[col].equals(update[col])


#################################
# PopulationView.update helpers #
##################################################
# PopulationView._update_column_and_ensure_dtype #
##################################################


def test__update_column_and_ensure_dtype() -> None:
    random.seed("test__update_column_and_ensure_dtype")

    for adding_simulants in [True, False]:
        for update_index in [BASE_POPULATION.index, BASE_POPULATION.index[::2]]:
            for col in BASE_POPULATION:
                update = pd.Series(
                    random.sample(BASE_POPULATION[col].tolist(), k=len(update_index)),
                    index=update_index,
                    name=col,
                )
                existing = BASE_POPULATION[col].copy()

                new_values = PopulationView._update_column_and_ensure_dtype(
                    update,
                    existing,
                    adding_simulants,
                )
                assert new_values.loc[update_index].astype(update.dtype).equals(update)
                non_update_index = existing.index.difference(update_index)
                if not non_update_index.empty:
                    assert new_values.loc[non_update_index].equals(
                        existing.loc[non_update_index]
                    )


def test__update_column_and_ensure_dtype_unmatched_dtype() -> None:
    # This tests a very specific failure case as the code is
    # not robust to general dtype silliness.
    update_index = BASE_POPULATION.index
    col = "count"
    update = pd.Series(
        random.sample(BASE_POPULATION[col].tolist(), k=len(update_index)),
        index=update_index,
        name=col,
    )
    existing = BASE_POPULATION[col].copy()
    # Count is an int, this coerces it to a float since there's no null type for ints.
    existing[:] = None

    # Should work fine when we're adding simulants
    new_values = PopulationView._update_column_and_ensure_dtype(
        update,
        existing,
        adding_simulants=True,
    )
    assert new_values.loc[update_index].equals(update)

    # And be bad news otherwise.
    with pytest.raises(PopulationError, match="corrupting"):
        PopulationView._update_column_and_ensure_dtype(
            update,
            existing,
            adding_simulants=False,
        )


#########################
# PopulationView.update #
#########################


def test_population_view_update_format_fail(
    population_manager: PopulationManager,
    population_update: pd.Series[Any] | pd.DataFrame,
    update_index: pd.Index[int],
) -> None:
    pv = population_manager.get_view(COL_NAMES)

    population_manager.creating_initial_population = True
    population_manager.adding_simulants = True
    # Bad type
    with pytest.raises(TypeError):
        pv.update(BASE_POPULATION.iloc[:, 0].tolist())  # type: ignore[arg-type]

    # Unknown population index
    if not update_index.empty:
        update = population_update.copy()
        update.index += 2 * update.index.max()
        with pytest.raises(PopulationError, match=f"{len(update)} simulants"):
            pv.update(update)

    # Missing population
    population_manager._population = BASE_POPULATION.loc[update_index]
    if not update_index.empty:
        with pytest.raises(PopulationError, match="missing updates"):
            pv.update(population_update.loc[update_index[::2]])

    # No new columns
    with pytest.raises(PopulationError, match="all provided columns"):
        pv.update(population_update)

    population_manager.creating_initial_population = False
    update = population_update.copy()
    if isinstance(update, pd.Series):
        update[:] = "bad_value"
    else:
        update.loc[:, COL_NAMES[0]] = "bad_value"
    if not update_index.empty:
        with pytest.raises(PopulationError, match="conflicting"):
            pv.update(update)


def test_population_view_update_format_fail_new_cols(
    population_manager: PopulationManager,
    population_update_new_cols: pd.Series[Any] | pd.DataFrame,
    update_index: pd.Index[int],
) -> None:
    pv = population_manager.get_view(COL_NAMES)

    population_manager.creating_initial_population = True
    population_manager.adding_simulants = True

    with pytest.raises(PopulationError, match="unnamed pandas series"):
        pv.update(BASE_POPULATION.iloc[:, 0].rename(None))

    for view_cols in [COL_NAMES, [COL_NAMES[0]]]:
        pv = population_manager.get_view(view_cols)

        with pytest.raises(PopulationError, match="extra columns"):
            pv.update(population_update_new_cols)

        with pytest.raises(PopulationError, match="no columns"):
            pv.update(pd.DataFrame(index=BASE_POPULATION.index))

    pv = population_manager.get_view(COL_NAMES + NEW_COL_NAMES)
    if not update_index.equals(BASE_POPULATION.index):
        with pytest.raises(PopulationError, match="missing updates"):
            pv.update(population_update_new_cols)

    # Conflicting data in existing cols.
    population_manager._population = BASE_POPULATION.loc[update_index]
    cols_overlap = [c for c in population_update_new_cols if c in COL_NAMES]
    if not update_index.empty and cols_overlap:
        update = population_update_new_cols.copy()
        update[COL_NAMES[0]] = "bad_values"
        with pytest.raises(PopulationError, match="conflicting"):
            pv.update(update)

    population_manager.creating_initial_population = False
    for adding_simulants in [True, False]:
        population_manager.adding_simulants = adding_simulants
        with pytest.raises(PopulationError, match="outside the initial population creation"):
            pv.update(population_update_new_cols)


def test_population_view_update_init(
    population_manager: PopulationManager,
    population_update_new_cols: pd.Series[Any] | pd.DataFrame,
    update_index: pd.Index[int],
) -> None:
    if isinstance(population_update_new_cols, pd.Series):
        pytest.skip()

    pv = population_manager.get_view(COL_NAMES + NEW_COL_NAMES)

    population_manager._population = BASE_POPULATION.loc[update_index]
    population_manager.creating_initial_population = True
    population_manager.adding_simulants = True
    pv.update(population_update_new_cols)

    for col in population_update_new_cols:
        assert population_manager._population[col].equals(population_update_new_cols[col])


def test_population_view_update_add(
    population_manager: PopulationManager,
    population_update: pd.Series[Any] | pd.DataFrame,
    update_index: pd.Index[int],
) -> None:
    if isinstance(population_update, pd.Series):
        pytest.skip()

    pv = population_manager.get_view(COL_NAMES + NEW_COL_NAMES)

    population_manager._population = BASE_POPULATION.loc[update_index]
    for col in population_update:
        population_manager._population[col] = None
    population_manager.creating_initial_population = False
    population_manager.adding_simulants = True
    pv.update(population_update)

    for col in population_update:
        if update_index.empty:
            assert population_manager._population[col].empty
        else:
            assert population_manager._population[col].equals(population_update[col])


def test_population_view_update_time_step(
    population_manager: PopulationManager,
    population_update: pd.Series[Any] | pd.DataFrame,
    update_index: pd.Index[int],
) -> None:
    if isinstance(population_update, pd.Series):
        pytest.skip()

    pv = population_manager.get_view(COL_NAMES + NEW_COL_NAMES)

    population_manager.creating_initial_population = False
    population_manager.adding_simulants = False
    pv.update(population_update)

    for col in population_update.columns:
        pop = population_manager._population
        assert pop is not None
        assert pop.loc[update_index, col].equals(population_update[col])



import numpy as np
import pandas as pd
import pytest

from vivarium.framework.utilities import (
    collapse_nested_dict,
    from_yearly,
    handle_exceptions,
    import_by_path,
    probability_to_rate,
    rate_to_probability,
    to_yearly,
)


def test_from_yearly() -> None:
    one_month = pd.Timedelta(days=30.5)
    rate = 0.01
    new_rate = from_yearly(rate, one_month)
    assert isinstance(new_rate, float)
    assert round(new_rate, 5) == round(0.0008356164383561645, 5)


def test_to_yearly() -> None:
    one_month = pd.Timedelta(days=30.5)
    rate = 0.0008356164383561645
    new_rate = to_yearly(rate, one_month)
    assert isinstance(new_rate, float)
    assert round(new_rate, 5) == round(0.01, 5)


def test_rate_to_probability() -> None:
    rate = np.array([0.001])
    prob = rate_to_probability(rate)
    assert np.isclose(prob, 0.00099950016662497809)


def test_very_high_rate_to_probability() -> None:
    rate = np.array([10_000])
    prob = rate_to_probability(rate)
    assert np.isclose(prob, 1.0)


def test_probability_to_rate() -> None:
    prob = np.array([0.00099950016662497809])
    rate = probability_to_rate(prob)
    assert np.isclose(rate, 0.001)


def test_rate_to_probability_symmetry() -> None:
    rate = np.array([0.0001])
    for _ in range(100):
        prob = rate_to_probability(rate)
        assert np.isclose(rate, probability_to_rate(prob))
        rate += (1 - 0.0001) / 100.0


def test_rate_to_probability_vectorizability() -> None:
    rate = 0.001
    rate_array = np.array([rate] * 100)
    prob_array = rate_to_probability(rate_array)
    sum_of_rates = np.sum(probability_to_rate(prob_array))

    assert isinstance(prob_array[10], float)
    assert round(prob_array[10], 5) == round(0.00099950016662497809, 5)
    assert isinstance(sum_of_rates, float)
    assert round(np.sum(rate_array), 5) == round(sum_of_rates, 5)


def test_collapse_nested_dict() -> None:
    source = {"a": {"b": {"c": 1, "d": 2}}, "e": 3}
    result = collapse_nested_dict(source)
    assert set(result) == {
        ("a.b.c", 1),
        ("a.b.d", 2),
        ("e", 3),
    }


def test_import_class_by_path() -> None:
    cls = import_by_path("collections.abc.Set")
    from collections.abc import Set

    assert cls is Set


def test_import_function_by_path() -> None:
    func = import_by_path("vivarium.framework.utilities.import_by_path")
    assert func is import_by_path


def test_bad_import_by_path() -> None:
    with pytest.raises(ImportError):
        import_by_path("junk.garbage.SillyClass")
    with pytest.raises(AttributeError):
        import_by_path("vivarium.framework.components.SillyClass")


class CustomException(Exception):
    pass


@pytest.mark.parametrize("test_input", [KeyboardInterrupt, RuntimeError, CustomException])
def test_handle_exceptions(test_input: type[BaseException]) -> None:
    def raise_me(ex: type[BaseException]) -> None:
        raise ex()

    with pytest.raises(test_input):
        # known issue with mypy
        # see heated thread at https://github.com/python/mypy/issues/6549
        func = handle_exceptions(raise_me(test_input), None, False)  # type: ignore[func-returns-value]
        func()



from __future__ import annotations

from collections.abc import Iterable
from datetime import datetime
from itertools import chain, combinations, product
from typing import Any

import numpy as np
import pandas as pd
import pytest
import pytest_mock
from scipy.stats import chisquare

from vivarium.framework.randomness import RandomnessError
from vivarium.framework.randomness.index_map import IndexMap


def almost_powerset(iterable: Iterable[int | str]) -> list[tuple[int | str, ...]]:
    """almost_powerset([1,2,3]) --> (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"""
    s = list(iterable)
    powerset = list(chain.from_iterable(combinations(s, r) for r in range(1, len(s) + 1)))
    return powerset


def generate_keys(
    number: int,
    types: tuple[str, str, str] = ("int", "float", "datetime"),
    seed: int = 123456,
) -> pd.DataFrame:
    rs = np.random.RandomState(seed=seed)

    keys: dict[str, Any] = {}
    if "datetime" in types:
        year = rs.choice(np.arange(1980, 2018))
        day = rs.choice(pd.date_range(f"01/01/{year}", periods=365))
        start_time = rs.choice(pd.date_range(day, periods=86400, freq="s"))
        freq = rs.choice(["ms", "s", "min", "h"])
        keys["datetime"] = pd.date_range(start_time, periods=number, freq=freq)

    if "int" in types:
        kind = rs.choice(["random", "sequential"])
        if kind == "random":
            ints = np.unique(rs.randint(0, 1000 * number, size=100 * number))
            assert len(ints) > number
            rs.shuffle(ints)
            keys["int"] = ints[:number]
        else:
            start = rs.randint(0, 100 * number)
            keys["int"] = np.arange(start, start + number, dtype=int)

    if "float" in types:
        keys["float"] = rs.random_sample(size=number)

    return pd.DataFrame(keys, index=pd.RangeIndex(number))


rs = np.random.RandomState(seed=456789)
index_sizes = list(rs.randint(10_000, 250_000, size=1))
types = list(almost_powerset(["int", "float", "datetime"]))
seeds = list(rs.randint(10000000, size=1))


def id_fun(param: tuple[float, Any, Any]) -> str:
    return f"Size:{param[0]}, Types:{param[1]}, Seed:{param[2]}"


@pytest.fixture(scope="module", params=list(product(index_sizes, types, seeds)), ids=id_fun)
def map_size_and_hashed_values(request: pytest.FixtureRequest) -> tuple[int, pd.Series[int]]:
    index_size, types_, seed = request.param
    keys = generate_keys(*request.param).set_index(types_).index
    m = IndexMap(key_columns=types_)
    return len(m), m._hash(keys)


def test_digit_series() -> None:
    m = IndexMap()
    k = pd.Series(123456789, index=range(10000))
    for i in range(10):
        assert len(m._digit(k, i).unique()) == 1
        assert m._digit(k, i)[0] == 10 - (i + 1)


def test_clip_to_seconds_scalar() -> None:
    m = IndexMap()
    k = pd.to_datetime("2010-01-25 06:25:31.123456789")
    assert (m._clip_to_seconds(pd.Series(k.value)) == int(str(k.value)[:10])).all()


def test_clip_to_seconds_series() -> None:
    m = IndexMap()
    stamp = 1234567890
    k = (
        pd.date_range(pd.to_datetime(stamp, unit="s"), periods=10000, freq="ns")
        .to_series()
        .astype(np.int64)
    )
    assert len(m._clip_to_seconds(k).unique()) == 1
    assert m._clip_to_seconds(k).unique()[0] == stamp


def test_spread_series() -> None:
    m = IndexMap()
    s = pd.Series(1234567890, index=range(10000))
    assert len(m._spread(s).unique()) == 1
    assert m._spread(s).unique()[0] == 4072825790


def test_shift_series() -> None:
    m = IndexMap()
    s = pd.Series(1.1234567890, index=range(10000))
    assert len(m._shift(s).unique()) == 1
    assert m._shift(s).unique()[0] == 1234567890


def test_convert_to_ten_digit_int() -> None:
    m = IndexMap()
    v = 1234567890
    date_col = pd.date_range(
        pd.to_datetime(v, unit="s"), periods=10000, freq="ns"
    ).to_series()
    int_col = pd.Series(v, index=range(10000))
    float_col = pd.Series(1.1234567890, index=range(10000))
    bad_col = pd.Series("a", index=range(10000))

    assert m._convert_to_ten_digit_int(date_col).unique()[0] == v
    assert m._convert_to_ten_digit_int(int_col).unique()[0] == 4072825790
    assert m._convert_to_ten_digit_int(float_col).unique()[0] == v
    with pytest.raises(RandomnessError):
        m._convert_to_ten_digit_int(bad_col)  # type: ignore [arg-type]


@pytest.mark.skip("This fails because the hash needs work")
def test_hash_collisions(map_size_and_hashed_values: tuple[int, pd.Series[int]]) -> None:
    n, h = map_size_and_hashed_values
    k = len(h)

    expected_empty_bins = n * (1 - 1 / n) ** k  # Potential source of roundoff issues.
    expected_full_bins = n - expected_empty_bins
    expected_collisions = k - expected_full_bins

    uniques = h.drop_duplicates()
    actual_collisions = len(h) - len(uniques)

    assert actual_collisions / expected_collisions < 1.5, "Too many collisions"


@pytest.mark.skip("This fails because the hash needs work")
def test_hash_uniformity(map_size_and_hashed_values: tuple[int, pd.Series[int]]) -> None:
    n, h = map_size_and_hashed_values

    k = len(h)
    num_bins = k // 5  # Want about 5 items per bin for chi-squared
    bins = pd.Series(np.linspace(0, n + 1, num_bins))

    binned_data = pd.cut(h, bins)
    distribution = pd.value_counts(binned_data).sort_index()
    c, p = chisquare(distribution)

    assert p > 0.05, "Data not uniform"


@pytest.fixture(scope="function")
def index_map(mocker: pytest_mock.MockFixture) -> type[IndexMap]:
    mock_index_map = IndexMap

    def hash_mock(
        k: pd.Index[int], salt: pd.Series[datetime] | pd.Series[int] | pd.Series[float]
    ) -> pd.Series[int]:
        seed = 123456
        digit_series = IndexMap()._convert_to_ten_digit_int(pd.Series(salt, index=k))
        rs = np.random.RandomState(seed=seed + digit_series)
        return pd.Series(rs.randint(0, len(k) * 10, size=len(k)), index=k)

    mocker.patch.object(mock_index_map, "_hash", side_effect=hash_mock)

    return mock_index_map


def test_update_empty_bad_keys(index_map: type[IndexMap]) -> None:
    keys = pd.DataFrame({"A": ["a"] * 10}, index=range(10))
    m = index_map(key_columns=list(keys.columns))
    with pytest.raises(RandomnessError):
        m.update(keys, pd.to_datetime("2023-01-01"))


def test_update_nonempty_bad_keys(index_map: type[IndexMap]) -> None:
    keys = generate_keys(1000)
    m = index_map(key_columns=list(keys.columns))
    m.update(keys, pd.to_datetime("2023-01-01"))
    with pytest.raises(RandomnessError):
        m.update(keys, pd.to_datetime("2023-01-01"))


def test_update_empty_good_keys(index_map: type[IndexMap]) -> None:
    keys = generate_keys(1000)
    m = index_map(key_columns=list(keys.columns))
    m.update(keys, pd.to_datetime("2023-01-01"))
    key_index = keys.set_index(list(keys.columns)).index
    map = m._map
    assert isinstance(map, pd.Series)

    assert len(map) == len(keys), "All keys not in mapping"
    assert (
        map.index.droplevel(m.SIM_INDEX_COLUMN).difference(key_index).empty
    ), "Extra keys in mapping"
    assert len(map.unique()) == len(keys), "Duplicate values in mapping"


def test_update_nonempty_good_keys(index_map: type[IndexMap]) -> None:
    keys = generate_keys(2000)
    m = index_map(key_columns=list(keys.columns))
    keys1, keys2 = keys[:1000], keys[1000:]

    m.update(keys1, pd.to_datetime("2023-01-01"))
    m.update(keys2, pd.to_datetime("2023-01-01"))
    map = m._map
    assert isinstance(map, pd.Series)

    key_index = keys.set_index(list(keys.columns)).index
    assert len(map) == len(keys), "All keys not in mapping"
    assert (
        map.index.droplevel(m.SIM_INDEX_COLUMN).difference(key_index).empty
    ), "Extra keys in mapping"
    assert len(map.unique()) == len(keys), "Duplicate values in mapping"



from __future__ import annotations

import itertools

import numpy as np
import pandas as pd

from vivarium import Component
from vivarium.framework.engine import Builder
from vivarium.framework.event import Event
from vivarium.framework.population import SimulantData
from vivarium.framework.results import VALUE_COLUMN
from vivarium.framework.results.observer import Observer
from vivarium.framework.results.stratification import Stratification
from vivarium.framework.values import Pipeline
from vivarium.types import ScalarMapper, VectorMapper

NAME = "hogwarts_house"
NAME_COLUMNS = ["first_name", "last_name"]
HOUSE_CATEGORIES = ["hufflepuff", "ravenclaw", "slytherin", "gryffindor"]
STUDENT_TABLE = pd.DataFrame(
    np.array([["harry", "potter"], ["severus", "snape"], ["luna", "lovegood"]]),
    columns=NAME_COLUMNS,
)
STUDENT_HOUSES = pd.Series(["gryffindor", "slytherin", "ravenclaw"])

BIN_BINNED_COLUMN = "silly_bin"
BIN_SOURCE = "silly_level"
BIN_LABELS = ["meh", "somewhat", "very", "extra"]
BIN_SILLY_BIN_EDGES = [0, 20, 40, 60, 90]

COL_NAMES = ["house", "familiar", "power_level", "tracked"]
FAMILIARS = ["owl", "cat", "gecko", "banana_slug", "unladen_swallow"]
POWER_LEVELS = [20, 40, 60, 80]
POWER_LEVEL_BIN_EDGES = [0, 25, 50, 75, 100]
POWER_LEVEL_GROUP_LABELS = ["low", "medium", "high", "very high"]
TRACKED_STATUSES = [True, False]
RECORDS = list(itertools.product(HOUSE_CATEGORIES, FAMILIARS, POWER_LEVELS, TRACKED_STATUSES))
BASE_POPULATION = pd.DataFrame(data=RECORDS, columns=COL_NAMES)

HARRY_POTTER_CONFIG = {
    "time": {
        "start": {"year": 2024, "month": 4, "day": 22},
        "end": {"year": 2029, "month": 4, "day": 22},
        "step_size": 365,  # Days
    },
    "stratification": {
        "default": ["student_house", "power_level_group"],
    },
}

RNG = np.random.default_rng(42)


##################
# Helper classes #
##################


class Hogwarts(Component):
    @property
    def columns_created(self) -> list[str]:
        return [
            "student_id",
            "student_house",
            "familiar",
            "power_level",
            "house_points",
            "quidditch_wins",
            "exam_score",
            "spell_power",
            "potion_power",
        ]

    def on_initialize_simulants(self, pop_data: SimulantData) -> None:
        size = len(pop_data.index)
        initialization_data = pd.DataFrame(
            {
                "student_id": list(range(size)),
                "student_house": RNG.choice(STUDENT_HOUSES, size=size),
                "familiar": RNG.choice(FAMILIARS, size=size),
                "power_level": RNG.choice([lvl for lvl in POWER_LEVELS], size=size),
                "house_points": 0,
                "quidditch_wins": 0,
                "exam_score": 0.0,
            },
            index=pop_data.index,
        )
        # Assume power level components are evenly split
        initialization_data["spell_power"] = initialization_data["power_level"] / 2
        initialization_data["potion_power"] = initialization_data["power_level"] / 2
        self.population_view.update(initialization_data)

    def on_time_step(self, pop_data: Event) -> None:
        update = self.population_view.get(pop_data.index)
        update["house_points"] = 0
        update["quidditch_wins"] = 0
        # House points are stratified by 'student_house' and 'power_level_group'.
        # Let's have each wizard of gryffindor and of power level 20 or 80
        # gain a point on each time step.
        update.loc[
            (update["student_house"] == "gryffindor")
            & (update["power_level"].isin([20, 80])),
            "house_points",
        ] = 1
        # Quidditch wins are stratified by 'familiar'.
        # Let's have each wizard with a banana slug familiar gain a point
        # on each time step.
        update.loc[update["familiar"] == "banana_slug", "quidditch_wins"] = 1
        # Update everyones test score to increase by 10 points per time step
        update["exam_score"] += 10.0

        self.population_view.update(update)


class HousePointsObserver(Observer):
    """Observer that is stratified by multiple columns (the defaults,
    'student_house' and 'power_level_group')
    """

    def register_observations(self, builder: Builder) -> None:
        builder.results.register_adding_observation(
            name="house_points",
            aggregator_sources=["house_points"],
            aggregator=lambda df: df.sum(),
            requires_columns=[
                "house_points",
            ],
            results_formatter=results_formatter,
        )


class FullyFilteredHousePointsObserver(Observer):
    """Same as `HousePointsObserver but with a filter that leaves no simulants"""

    def register_observations(self, builder: Builder) -> None:
        builder.results.register_adding_observation(
            name="house_points",
            pop_filter="tracked==True & power_level=='one billion'",
            aggregator_sources=["house_points"],
            aggregator=lambda df: df.sum(),
            requires_columns=[
                "house_points",
            ],
        )


class QuidditchWinsObserver(Observer):
    """Observer that is stratified by a single column ('familiar')"""

    def register_observations(self, builder: Builder) -> None:
        builder.results.register_adding_observation(
            name="quidditch_wins",
            aggregator_sources=["quidditch_wins"],
            aggregator=lambda df: df.sum(),
            excluded_stratifications=["student_house", "power_level_group"],
            additional_stratifications=["familiar"],
            requires_columns=[
                "quidditch_wins",
            ],
            results_formatter=results_formatter,
        )


class NoStratificationsQuidditchWinsObserver(Observer):
    """Same as above but no stratifications at all"""

    def register_observations(self, builder: Builder) -> None:
        builder.results.register_adding_observation(
            name="no_stratifications_quidditch_wins",
            aggregator_sources=["quidditch_wins"],
            aggregator=lambda df: df.sum(),
            excluded_stratifications=["student_house", "power_level_group"],
            requires_columns=[
                "quidditch_wins",
            ],
            results_formatter=results_formatter,
        )


class MagicalAttributesObserver(Observer):
    """Observer whose aggregator returns a pd.Series instead of a float (which in
    turn results in a dataframe with multiple columns instead of just one
    'value' column)
    """

    def register_observations(self, builder: Builder) -> None:
        builder.results.register_adding_observation(
            name="magical_attributes",
            aggregator=self._get_magical_attributes,
            excluded_stratifications=["student_house"],
            results_formatter=lambda *_: pd.DataFrame(),
        )

    def _get_magical_attributes(self, _: pd.DataFrame) -> pd.Series[float]:
        """Increase each level by one per time step"""
        return pd.Series([1.0, 1.0], ["spell_power", "potion_power"])


class ExamScoreObserver(Observer):
    """Observer that is not stratified and exam scores"""

    def register_observations(self, builder: Builder) -> None:
        builder.results.register_concatenating_observation(
            name="exam_score",
            requires_columns=["student_id", "student_house", "exam_score"],
        )


class CatBombObserver(Observer):
    """Observer that counts the number of feral cats per house"""

    def register_observations(self, builder: Builder) -> None:
        builder.results.register_stratified_observation(
            name="cat_bomb",
            pop_filter="familiar=='cat' and tracked==True",
            requires_columns=["familiar"],
            results_updater=self.update_cats,
            excluded_stratifications=["power_level_group"],
            aggregator_sources=["student_house"],
            aggregator=len,
        )

    def update_cats(self, existing_df: pd.DataFrame, new_df: pd.DataFrame) -> pd.DataFrame:
        no_cats_mask = existing_df["value"] == 0
        updated_df = existing_df
        updated_df.loc[no_cats_mask, "value"] = new_df["value"]
        updated_df.loc[~no_cats_mask, "value"] *= new_df["value"]
        return updated_df


class ValedictorianObserver(Observer):
    """Observer that records the valedictorian at each time step. All students
    have the same exam scores and so the valedictorian is chosen randomly.
    """

    def __init__(self) -> None:
        super().__init__()
        self.valedictorians: list[int] = []

    def register_observations(self, builder: Builder) -> None:
        builder.results.register_unstratified_observation(
            name="valedictorian",
            requires_columns=["event_time", "student_id", "exam_score"],
            results_gatherer=self.choose_valedictorian,  # type: ignore [arg-type]
            results_updater=self.update_valedictorian,
        )

    def choose_valedictorian(self, df: pd.DataFrame) -> pd.DataFrame:
        eligible_students = df.loc[~df["student_id"].isin(self.valedictorians), "student_id"]
        valedictorian: int = RNG.choice(eligible_students)
        self.valedictorians.append(valedictorian)
        return df[df["student_id"] == valedictorian]

    def update_valedictorian(
        self, _existing_df: pd.DataFrame, new_df: pd.DataFrame
    ) -> pd.DataFrame:
        return new_df


class HogwartsResultsStratifier(Component):
    def setup(self, builder: Builder) -> None:
        builder.results.register_stratification(
            name="student_house",
            categories=list(STUDENT_HOUSES),
            requires_columns=["student_house"],
        )
        builder.results.register_stratification(
            name="familiar", categories=FAMILIARS, requires_columns=["familiar"]
        )
        builder.results.register_binned_stratification(
            "power_level",
            "power_level_group",
            POWER_LEVEL_BIN_EDGES,
            POWER_LEVEL_GROUP_LABELS,
        )


##################
# Helper methods #
##################


def results_formatter(
    measure: str,
    results: pd.DataFrame,
) -> pd.DataFrame:
    """An test use case of an observer's report method that writes a DataFrame to a CSV file."""
    # Add extra cols
    results["measure"] = measure
    # Sort the columns such that the stratifications (index) are first
    # and VALUE_COLUMN is last and sort the rows by the stratifications.
    other_cols = [c for c in results.columns if c != VALUE_COLUMN]
    return results[other_cols + [VALUE_COLUMN]].sort_index().reset_index()


def sorting_hat_vectorized(state_table: pd.DataFrame) -> pd.Series[str]:
    sorted_series = state_table.apply(sorting_hat_serial, axis=1)
    return sorted_series


def sorting_hat_serial(simulant_row: pd.Series[str]) -> str:
    first_name = simulant_row[0]
    last_name = simulant_row[1]
    if first_name == "harry":
        return "gryffindor"
    if first_name == "luna":
        return "ravenclaw"
    if last_name == "snape":
        return "slytherin"
    return "hufflepuff"


def sorting_hat_bad_mapping(simulant_row: pd.Series[str]) -> str:
    # Return something not in CATEGORIES
    return "pancakes"


def verify_stratification_added(
    stratification_list: list[Stratification],
    name: str,
    sources: list[str],
    categories: list[str],
    excluded_categories: list[str],
    mapper: VectorMapper | ScalarMapper,
    is_vectorized: bool,
) -> bool:
    """Verify that a Stratification object is in `stratification_list`"""
    matching_stratification_found = False
    for stratification in stratification_list:  # noqa
        # big equality check
        if (
            stratification.name == name
            and sorted(stratification.categories)
            == sorted([cat for cat in categories if cat not in excluded_categories])
            and sorted(stratification.excluded_categories) == sorted(excluded_categories)
            and stratification.mapper == mapper
            and stratification.is_vectorized == is_vectorized
            and sorted(stratification.sources) == sorted(sources)
        ):
            matching_stratification_found = True
            break
    return matching_stratification_found


# Mock for get_value call for Pipelines, returns a str instead of a Pipeline
def mock_get_value(self: Builder, name: str) -> Pipeline:
    if not isinstance(name, str):
        raise TypeError("Passed a non-string type to mock get_value(), check your pipelines.")
    return Pipeline(name)



"""
============================
Randomness System Exceptions
============================

Errors related to improper use of the Randomness system.

"""

from vivarium.exceptions import VivariumError


class RandomnessError(VivariumError):
    """Raised for inconsistencies in random number and choice generation."""

    pass



from collections.abc import Callable, Mapping
from datetime import datetime, timedelta
from typing import TYPE_CHECKING
from typing import SupportsFloat as Numeric
from typing import Union

import numpy as np
import numpy.typing as npt
import pandas as pd

if TYPE_CHECKING:
    from vivarium.framework.engine import Builder

NumericArray = npt.NDArray[np.number[npt.NBitBase]]

Time = pd.Timestamp | datetime
Timedelta = pd.Timedelta | timedelta
ClockTime = Time | int
ClockStepSize = Timedelta | int

ScalarValue = Numeric | Timedelta | Time
LookupTableData = (
    ScalarValue
    | pd.DataFrame
    | list[ScalarValue]
    | tuple[ScalarValue, ...]
    | Mapping[str, list[ScalarValue] | list[str]]
)

DataInput = LookupTableData | str | Callable[["Builder"], LookupTableData]

# TODO: For some of the uses of NumberLike, we probably want a TypeVar here instead.
NumberLike = Union[
    NumericArray,
    # TODO: Parameterizing pandas objects fails below python 3.12
    pd.Series,  # type: ignore [type-arg]
    pd.DataFrame,
    float,
    int,
]

VectorMapper = Callable[[pd.DataFrame], pd.Series]  # type: ignore [type-arg]
ScalarMapper = Callable[[pd.Series], str]  # type: ignore [type-arg]



from __future__ import annotations

from datetime import timedelta
from typing import TYPE_CHECKING, Any, Protocol

import pandas as pd

from vivarium.framework.utilities import from_yearly
from vivarium.framework.values.exceptions import DynamicValueError
from vivarium.types import NumberLike

if TYPE_CHECKING:
    from vivarium.framework.values.manager import ValuesManager


class PostProcessor(Protocol):
    def __call__(self, value: Any, manager: ValuesManager) -> Any:
        ...


def rescale_post_processor(value: NumberLike, manager: ValuesManager) -> NumberLike:
    """Rescales annual rates to time-step appropriate rates.

    This should only be used with a simulation using a
    :class:`~vivarium.framework.time.DateTimeClock` or another implementation
    of a clock that traffics in pandas date-time objects.

    Parameters
    ----------
    value
        Annual rates, either as a number or something we can broadcast
        multiplication over like a :mod:`numpy` array or :mod:`pandas`
        data frame.
    manager
        The ValuesManager for this simulation.

    Returns
    -------
        The annual rates rescaled to the size of the current time step size.
    """
    if isinstance(value, (pd.Series, pd.DataFrame)):
        return value.mul(
            manager.simulant_step_sizes(value.index)
            .astype("timedelta64[ns]")
            .dt.total_seconds()
            / (60 * 60 * 24 * 365.0),
            axis=0,
        )
    else:
        time_step = manager.step_size()
        if not isinstance(time_step, (pd.Timedelta, timedelta)):
            raise DynamicValueError(
                "The rescale post processor requires a time step size that is a "
                "datetime timedelta or pandas Timedelta object."
            )
        return from_yearly(value, time_step)


def union_post_processor(values: list[NumberLike], _: Any) -> NumberLike:
    """Computes a probability on the union of the sample spaces in the values.

    Given a list of values where each value is a probability of an independent
    event, this post processor computes the probability of the union of the
    events.

    .. list-table::
       :width: 100%
       :widths: 1 3

       * - :math:`p_x`
         - Probability of event x
       * - :math:`1 - p_x`
         - Probability of not event x
       * - :math:`\prod_x(1 - p_x)`
         - Probability of not any events x
       * - :math:`1 - \prod_x(1 - p_x)`
         - Probability of any event x

    Parameters
    ----------
    values
        A list of independent proportions or probabilities, either
        as numbers or as a something we can broadcast addition and
        multiplication over.

    Returns
    -------
        The probability over the union of the sample spaces represented
        by the original probabilities.
    """
    # if there is only one value, return the value
    if len(values) == 1:
        return values[0]

    # if there are multiple values, calculate the joint value
    product: NumberLike = 1
    for v in values:
        new_value = 1 - v
        product = product * new_value
    joint_value = 1 - product
    return joint_value



"""
===================
The Vivarium Engine
===================

The engine houses the :class:`SimulationContext` -- the key ``vivarium`` object
for running and interacting with simulations. It is the top level manager
for all state information in ``vivarium``.  By intention, it exposes a very
simple interface for managing the
:ref:`simulation lifecycle <lifecycle_concept>`.

Also included here is the simulation :class:`Builder`, which is the main
interface that components use to interact with the simulation framework. You
can read more about how the builder works and what services is exposes
:ref:`here <builder_concept>`.

Finally, there are a handful of wrapper methods that allow a user or user
tools to easily setup and run a simulation.

"""

from pathlib import Path
from pprint import pformat
from time import time
from typing import Any

import dill
import numpy as np
import pandas as pd
from layered_config_tree.exceptions import ConfigurationKeyError
from layered_config_tree.main import LayeredConfigTree

from vivarium import Component
from vivarium.exceptions import VivariumError
from vivarium.framework.artifact import ArtifactInterface, ArtifactManager
from vivarium.framework.components import (
    ComponentConfigError,
    ComponentInterface,
    ComponentManager,
)
from vivarium.framework.configuration import build_model_specification
from vivarium.framework.event import EventInterface, EventManager
from vivarium.framework.lifecycle import LifeCycleInterface, LifeCycleManager
from vivarium.framework.logging import LoggingInterface, LoggingManager
from vivarium.framework.lookup import LookupTableInterface, LookupTableManager
from vivarium.framework.plugins import PluginManager
from vivarium.framework.population import PopulationInterface, PopulationManager
from vivarium.framework.randomness import RandomnessInterface, RandomnessManager
from vivarium.framework.resource import ResourceInterface, ResourceManager
from vivarium.framework.results import ResultsInterface, ResultsManager
from vivarium.framework.time import SimulationClock, TimeInterface
from vivarium.framework.values import ValuesInterface, ValuesManager
from vivarium.types import ClockTime


class SimulationContext:
    _created_simulation_contexts: set[str] = set()

    @staticmethod
    def _get_context_name(sim_name: str | None) -> str:
        """Get a unique name for a simulation context.

        Parameters
        ----------
        sim_name
            The name of the simulation context.  If None, a unique name will be generated.

        Returns
        -------
            A unique name for the simulation context.

        Notes
        -----
        This method mutates process global state (the class attribute
        ``_created_simulation_contexts``) in order to keep track contexts that have been
        generated. This functionality makes generating simulation contexts in parallel
        a non-threadsafe operation.
        """
        if sim_name is None:
            sim_number = len(SimulationContext._created_simulation_contexts) + 1
            sim_name = f"simulation_{sim_number}"

        if sim_name in SimulationContext._created_simulation_contexts:
            msg = (
                "Attempting to create two SimulationContexts "
                f"with the same name {sim_name}"
            )
            raise VivariumError(msg)

        SimulationContext._created_simulation_contexts.add(sim_name)
        return sim_name

    @staticmethod
    def _clear_context_cache() -> None:
        """Clear the cache of simulation context names.

        Notes
        -----
        This is primarily useful for testing purposes.
        """
        SimulationContext._created_simulation_contexts = set()

    def __init__(
        self,
        model_specification: str | Path | LayeredConfigTree | None = None,
        components: list[Component] | dict[str, Any] | LayeredConfigTree | None = None,
        configuration: dict[str, Any] | LayeredConfigTree | None = None,
        plugin_configuration: dict[str, Any] | LayeredConfigTree | None = None,
        sim_name: str | None = None,
        logging_verbosity: int = 1,
    ) -> None:
        self._name = self._get_context_name(sim_name)

        # Bootstrap phase: Parse arguments, make private managers
        component_configuration = (
            components if isinstance(components, (dict, LayeredConfigTree)) else None
        )
        self._additional_components = components if isinstance(components, list) else []
        self.model_specification = build_model_specification(
            model_specification, component_configuration, configuration, plugin_configuration
        )

        self._plugin_configuration = self.model_specification.plugins
        self._component_configuration = self.model_specification.components
        self.configuration = self.model_specification.configuration

        self._plugin_manager = PluginManager(self.model_specification.plugins)

        self._logging = self._plugin_manager.get_plugin(LoggingManager)
        self._logging.configure_logging(
            simulation_name=self.name,
            verbosity=logging_verbosity,
        )
        self._logger = self._logging.get_logger()

        self._builder = Builder(self.configuration, self._plugin_manager)

        # This formally starts the initialization phase (this call makes the
        # life-cycle manager).
        self._lifecycle = self._plugin_manager.get_plugin(LifeCycleManager)
        self._lifecycle.add_phase("setup", ["setup", "post_setup", "population_creation"])
        self._lifecycle.add_phase(
            "main_loop",
            ["time_step__prepare", "time_step", "time_step__cleanup", "collect_metrics"],
            loop=True,
        )
        self._lifecycle.add_phase("simulation_end", ["simulation_end", "report"])

        self._component_manager = self._plugin_manager.get_plugin(ComponentManager)
        self._component_manager.setup_manager(self.configuration, self._lifecycle)

        self._clock = self._plugin_manager.get_plugin(SimulationClock)
        self._values = self._plugin_manager.get_plugin(ValuesManager)
        self._events = self._plugin_manager.get_plugin(EventManager)
        self._population = self._plugin_manager.get_plugin(PopulationManager)
        self._resource = self._plugin_manager.get_plugin(ResourceManager)
        self._results = self._plugin_manager.get_plugin(ResultsManager)
        self._tables = self._plugin_manager.get_plugin(LookupTableManager)
        self._randomness = self._plugin_manager.get_plugin(RandomnessManager)
        self._data = self._plugin_manager.get_plugin(ArtifactManager)

        optional_managers = self._plugin_manager.get_optional_controllers()
        for name in optional_managers:
            setattr(self, f"_{name}", optional_managers[name])

        # The order the managers are added is important.  It represents the
        # order in which they will be set up.  The logging manager and the clock are
        # required by several of the other managers, including the lifecycle manager. The
        # lifecycle manager is also required by most managers. The randomness
        # manager requires the population manager.  The remaining managers need
        # no ordering.
        managers = [
            self._logging,
            self._lifecycle,
            self._resource,
            self._values,
            self._population,
            self._clock,
            self._randomness,
            self._events,
            self._tables,
            self._data,
            self._results,
        ] + list(self._plugin_manager.get_optional_controllers().values())
        self._component_manager.add_managers(managers)

        component_config_parser = self._plugin_manager.get_component_config_parser()
        # Tack extra components onto the end of the list generated from the model specification.
        components_list: list[Component] = (
            component_config_parser.get_components(self._component_configuration)
            + self._additional_components
        )

        non_components = [obj for obj in components_list if not isinstance(obj, Component)]
        if non_components:
            message = (
                "Attempting to create a simulation with the following components "
                "that do not inherit from `vivarium.Component`: "
                f"[{[c.name for c in non_components]}]."
            )
            raise ComponentConfigError(message)

        self._lifecycle.add_constraint(self.add_components, allow_during=["initialization"])
        self._lifecycle.add_constraint(
            self.get_population, restrict_during=["initialization", "setup", "post_setup"]
        )

        self.add_components(components_list)

    @property
    def name(self) -> str:
        return self._name

    @property
    def current_time(self) -> ClockTime:
        """Returns the current simulation time."""
        return self._clock.time

    def get_results(self) -> dict[str, pd.DataFrame]:
        """Return the formatted results."""
        return self._results.get_results()

    def run_simulation(self) -> None:
        """A wrapper method to run all steps of a simulation"""
        self.setup()
        self.initialize_simulants()
        self.run()
        self.finalize()
        self.report()

    def setup(self) -> None:
        self._lifecycle.set_state("setup")
        self.configuration.freeze()
        self._component_manager.setup_components(self._builder)

        self.simulant_creator = self._builder.population.get_simulant_creator()

        self.time_step_events = self._lifecycle.get_state_names("main_loop")
        self.time_step_emitters = {
            k: self._builder.event.get_emitter(k) for k in self.time_step_events
        }
        self.end_emitter = self._builder.event.get_emitter("simulation_end")
        self.report_emitter = self._builder.event.get_emitter("report")

        post_setup = self._builder.event.get_emitter("post_setup")
        self._lifecycle.set_state("post_setup")
        post_setup(pd.Index([]), None)

    def initialize_simulants(self) -> None:
        self._lifecycle.set_state("population_creation")
        pop_params = self.configuration.population
        # Fencepost the creation of the initial population.
        self._clock.step_backward()
        population_size = pop_params.population_size
        self.simulant_creator(population_size, {"sim_state": "setup"})
        self._clock.step_forward(self.get_population().index)

    def step(self) -> None:
        self._logger.debug(self.current_time)
        for event in self.time_step_events:
            self._logger.debug(f"Event: {event}")
            self._lifecycle.set_state(event)
            pop_to_update = self._clock.get_active_simulants(
                self.get_population().index,
                self._clock.event_time,
            )
            self._logger.debug(f"Updating: {len(pop_to_update)}")
            self.time_step_emitters[event](pop_to_update, None)
        self._clock.step_forward(self.get_population().index)

    def run(
        self,
        backup_path: Path | None = None,
        backup_freq: int | float | None = None,
    ) -> None:
        if backup_freq and backup_path:
            time_to_save = time() + backup_freq
            while self.current_time < self._clock.stop_time:  # type: ignore [operator]
                self.step()
                if time() >= time_to_save:
                    self._logger.debug(f"Writing Simulation Backup to {backup_path}")
                    self.write_backup(backup_path)
                    time_to_save = time() + backup_freq
        else:
            while self.current_time < self._clock.stop_time:  # type: ignore [operator]
                self.step()

    def finalize(self) -> None:
        self._lifecycle.set_state("simulation_end")
        self.end_emitter(self.get_population().index, None)
        unused_config_keys = self.configuration.unused_keys()
        if unused_config_keys:
            self._logger.warning(
                f"Some configuration keys not used during run: {unused_config_keys}."
            )

    def report(self, print_results: bool = True) -> None:
        self._lifecycle.set_state("report")
        self.report_emitter(self.get_population().index, None)
        results = self.get_results()
        if print_results:
            for measure, df in results.items():
                self._logger.info(f"\n{measure}:\n{pformat(df)}")
            performance_metrics = self.get_performance_metrics()
            performance_metrics_str: str = performance_metrics.to_string(
                index=False,
                float_format=lambda x: f"{x:.2f}",
            )
            self._logger.info("\n" + performance_metrics_str)
        self._write_results(results)

    def _write_results(self, results: dict[str, pd.DataFrame]) -> None:
        """Iterate through the measures and write out the formatted results"""
        try:
            results_dir = self.configuration.output_data.results_directory
            for measure, df in results.items():
                output_file = Path(results_dir) / f"{measure}.parquet"
                df.to_parquet(output_file, index=False)
        except ConfigurationKeyError:
            self._logger.info("No results directory set; results are not written to disk.")

    def write_backup(self, backup_path: Path) -> None:
        with open(backup_path, "wb") as f:
            dill.dump(self, f, protocol=dill.HIGHEST_PROTOCOL)

    def get_performance_metrics(self) -> pd.DataFrame:
        timing_dict = self._lifecycle.timings
        total_time = np.sum([np.sum(v) for v in timing_dict.values()])
        timing_dict["total"] = [total_time]
        records = [
            {
                "Event": label,
                "Count": len(ts),
                "Mean time (s)": np.mean(ts),
                "Std. dev. time (s)": np.std(ts),
                "Total time (s)": sum(ts),
                "% Total time": 100 * sum(ts) / total_time,
            }
            for label, ts in timing_dict.items()
        ]
        performance_metrics = pd.DataFrame(records)
        return performance_metrics

    def add_components(self, component_list: list[Component]) -> None:
        """Adds new components to the simulation."""
        self._component_manager.add_components(component_list)

    def get_population(self, untracked: bool = True) -> pd.DataFrame:
        return self._population.get_population(untracked)

    def __repr__(self) -> str:
        return f"SimulationContext({self.name})"

    def get_number_of_steps_remaining(self) -> int:
        return self._clock.time_steps_remaining

    @classmethod
    def load_from_backup(cls, backup_path: Path) -> "SimulationContext":
        """Load a simulation context from a backup file."""
        with open(backup_path, "rb") as f:
            backup: SimulationContext = dill.load(f)
        return backup


class Builder:
    """Toolbox for constructing and configuring simulation components.

    This is the access point for components through which they are able to
    utilize a variety of interfaces to interact with the simulation framework.

    Notes
    -----
    A `Builder` should never be created directly. It will automatically be
    created during the initialization of a :class:`SimulationContext`

    """

    def __init__(
        self, configuration: LayeredConfigTree, plugin_manager: PluginManager
    ) -> None:
        self.configuration = configuration
        """Provides access to the :ref:`configuration<configuration_concept>`"""

        self.logging = plugin_manager.get_plugin_interface(LoggingInterface)
        """Provides access to the :ref:`logging<logging_concept>` system."""

        self.lookup = plugin_manager.get_plugin_interface(LookupTableInterface)
        """Provides access to simulant-specific data via the
        :ref:`lookup table<lookup_concept>` abstraction."""

        self.value = plugin_manager.get_plugin_interface(ValuesInterface)
        """Provides access to computed simulant attribute values via the
        :ref:`value pipeline<values_concept>` system."""

        self.event = plugin_manager.get_plugin_interface(EventInterface)
        """Provides access to event listeners utilized in the
        :ref:`event<event_concept>` system."""

        self.population = plugin_manager.get_plugin_interface(PopulationInterface)
        """Provides access to simulant state table via the
        :ref:`population<population_concept>` system."""

        self.resources = plugin_manager.get_plugin_interface(ResourceInterface)
        """Provides access to the :ref:`resource<resource_concept>` system,
         which manages dependencies between components.
         """

        self.results = plugin_manager.get_plugin_interface(ResultsInterface)
        """Provides access to the :ref:`results<results_concept>` system."""

        self.randomness = plugin_manager.get_plugin_interface(RandomnessInterface)
        """Provides access to the :ref:`randomness<crn_concept>` system."""

        self.time: TimeInterface = plugin_manager.get_plugin_interface(TimeInterface)
        """Provides access to the simulation's :ref:`clock<time_concept>`."""

        self.components = plugin_manager.get_plugin_interface(ComponentInterface)
        """Provides access to the :ref:`component management<components_concept>`
        system, which maintains a reference to all managers and components in
        the simulation."""

        self.lifecycle = plugin_manager.get_plugin_interface(LifeCycleInterface)
        """Provides access to the :ref:`life-cycle<lifecycle_concept>` system,
        which manages the simulation's execution life-cycle."""

        self.data = plugin_manager.get_plugin_interface(ArtifactInterface)
        """Provides access to the simulation's input data housed in the
        :ref:`data artifact<data_concept>`."""

        for name, interface in plugin_manager.get_optional_interfaces().items():
            setattr(self, name, interface)

    def __repr__(self) -> str:
        return "Builder()"



"""
================
Resource Manager
================

"""

from __future__ import annotations

from collections.abc import Iterable
from typing import TYPE_CHECKING, Any

import networkx as nx

from vivarium.framework.resource.exceptions import ResourceError
from vivarium.framework.resource.group import ResourceGroup
from vivarium.framework.resource.resource import Column, NullResource, Resource
from vivarium.manager import Interface, Manager

if TYPE_CHECKING:
    from vivarium import Component
    from vivarium.framework.engine import Builder


class ResourceManager(Manager):
    """Manages all the resources needed for population initialization."""

    def __init__(self) -> None:
        self._resource_group_map: dict[str, ResourceGroup] = {}
        """Resource - resource group mapping. This will be a dict with string keys 
        representing the the resource and the resource group they belong to. This 
        is a one to many mapping as some resource groups contain many resources."""
        self._null_producer_count = 0
        """Null producer counter. Null producers are those that don't produce any 
        resources externally but still consume other resources (i.e., have 
        dependencies) - these are only pop initializers as of 9/26/2019. Tracker 
        is here to assign them unique ids."""
        self._graph: nx.DiGraph | None = None
        """Attribute used for lazy (but cached) graph initialization."""
        self._sorted_nodes: list[ResourceGroup] | None = None
        """Attribute used for lazy (but cached) graph topo sort."""

    @property
    def name(self) -> str:
        """The name of this manager."""
        return "resource_manager"

    @property
    def graph(self) -> nx.DiGraph:
        """The networkx graph representation of the resource pool."""
        if self._graph is None:
            self._graph = self._to_graph()
        return self._graph

    @property
    def sorted_nodes(self) -> list[ResourceGroup]:
        """Returns a topological sort of the resource graph.

        Notes
        -----
        Topological sorts are not stable. Be wary of depending on order
        where you shouldn't.
        """
        if self._sorted_nodes is None:
            try:
                self._sorted_nodes = list(nx.algorithms.topological_sort(self.graph))  # type: ignore[func-returns-value]
            except nx.NetworkXUnfeasible:
                raise ResourceError(
                    "The resource pool contains at least one cycle: "
                    f"{nx.find_cycle(self.graph)}."
                )
        return self._sorted_nodes

    def setup(self, builder: Builder) -> None:
        self.logger = builder.logging.get_logger(self.name)

    def add_resources(
        self,
        # TODO [MIC-5452]: all resource groups should have a component
        component: Component | Manager | None,
        resources: Iterable[str | Resource],
        dependencies: Iterable[str | Resource],
    ) -> None:
        """Adds managed resources to the resource pool.

        Parameters
        ----------
        component
            The component or manager adding the resources.
        resources
            The resources being added. A string represents a column resource.
        dependencies
            A list of resources that the producer requires. A string represents
            a column resource.

        Raises
        ------
        ResourceError
            If a component has multiple resource producers for the ``column``
            resource type or there are multiple producers of the same resource.
        """
        resource_group = self._get_resource_group(component, resources, dependencies)

        for resource_id, resource in resource_group.resources.items():
            if resource_id in self._resource_group_map:
                other_resource = self._resource_group_map[resource_id]
                # TODO [MIC-5452]: all resource groups should have a component
                resource_component = resource.component.name if resource.component else None
                other_resource_component = (
                    other_resource.component.name if other_resource.component else None
                )
                raise ResourceError(
                    f"Component '{resource_component}' is attempting to register"
                    f" resource '{resource_id}' but it is already registered by"
                    f" '{other_resource_component}'."
                )
            self._resource_group_map[resource_id] = resource_group

    def _get_resource_group(
        self,
        component: Component | Manager | None,
        resources: Iterable[str | Resource],
        dependencies: Iterable[str | Resource],
    ) -> ResourceGroup:
        """Packages resource information into a resource group.

        See Also
        --------
        :class:`ResourceGroup`
        """
        resources_ = [Column(r, component) if isinstance(r, str) else r for r in resources]
        dependencies_ = [Column(d, None) if isinstance(d, str) else d for d in dependencies]

        if not resources_:
            # We have a "producer" that doesn't produce anything, but
            # does have dependencies. This is necessary for components that
            # want to track private state information.
            resources_ = [NullResource(self._null_producer_count, component)]
            self._null_producer_count += 1

        # TODO [MIC-5452]: all resource groups should have a component
        if component and (
            have_other_component := [r for r in resources_ if r.component != component]
        ):
            raise ResourceError(
                f"All initialized resources must have the component '{component.name}'."
                f" The following resources have a different component: {have_other_component}"
            )

        return ResourceGroup(resources_, dependencies_)

    def _to_graph(self) -> nx.DiGraph:
        """Constructs the full resource graph from information in the groups.

        Components specify local dependency information during setup time.
        When the resources are required at population creation time,
        the graph is generated as all resources must be registered at that
        point.

        Notes
        -----
        We are taking advantage of lazy initialization to sneak this in
        between post setup time when the :class:`values manager
        <vivarium.framework.values.ValuesManager>` finalizes pipeline
        dependencies and population creation time.
        """
        resource_graph = nx.DiGraph()
        # networkx ignores duplicates
        resource_graph.add_nodes_from(self._resource_group_map.values())

        for resource_group in resource_graph.nodes:
            for dependency in resource_group.dependencies:
                if dependency not in self._resource_group_map:
                    # Warn here because this sometimes happens naturally
                    # if observer components are missing from a simulation.
                    self.logger.warning(
                        f"Resource {dependency} is not produced by any"
                        f" component but is needed to compute {resource_group}."
                    )
                    continue
                dependency_group = self._resource_group_map[dependency]
                resource_graph.add_edge(dependency_group, resource_group)

        return resource_graph

    def get_population_initializers(self) -> list[Any]:
        """Returns a dependency-sorted list of population initializers.

        We exclude all non-initializer dependencies. They were necessary in
        graph construction, but we only need the column producers at population
        creation time.
        """
        return [r.initializer for r in self.sorted_nodes if r.is_initialized]

    def __repr__(self) -> str:
        out = {}
        for resource_group in set(self._resource_group_map.values()):
            produced = ", ".join(resource_group)
            out[produced] = ", ".join(resource_group.dependencies)
        return "\n".join([f"{produced} : {depends}" for produced, depends in out.items()])


class ResourceInterface(Interface):
    """The resource management system.

    A resource in :mod:`vivarium` is something like a state table column
    or a randomness stream. These resources are used to initialize or alter
    the state of the simulation. Many of these resources might depend on each
    other and therefore need to be created or updated in a particular order.
    These dependency chains can be quite long and complex.

    Placing the ordering responsibility on end users makes simulations very
    fragile and difficult to understand. Instead, the resource management
    system allows users to only specify local dependencies. The system then
    uses the local dependency information to construct a full dependency
    graph, validate that there are no cyclic dependencies, and return
    resources and their producers in an order that makes sense.

    """

    def __init__(self, manager: ResourceManager):
        self._manager = manager

    def add_resources(
        self,
        # TODO [MIC-5452]: all resource groups should have a component
        component: Component | Manager | None,
        resources: Iterable[str | Resource],
        dependencies: Iterable[str | Resource],
    ) -> None:
        """Adds managed resources to the resource pool.

        Parameters
        ----------
        component
            The component or manager adding the resources.
        resources
            The resources being added. A string represents a column resource.
        dependencies
            A list of resources that the producer requires. A string represents
            a column resource.

        Raises
        ------
        ResourceError
            If either the resource type is invalid, a component has multiple
            resource producers for the ``column`` resource type, or
            there are multiple producers of the same resource.
        """
        self._manager.add_resources(component, resources, dependencies)

    def get_population_initializers(self) -> list[Any]:
        """Returns a dependency-sorted list of population initializers.

        We exclude all non-initializer dependencies. They were necessary in
        graph construction, but we only need the column producers at population
        creation time.
        """
        return self._manager.get_population_initializers()



import numpy as np
import pandas as pd

from vivarium import Component
from vivarium.framework.engine import Builder
from vivarium.framework.population import SimulantData


class Population(Component):
    ##############
    # Properties #
    ##############
    configuration_defaults = {
        "population": {
            "colors": ["red", "blue"],
        }
    }
    columns_created = ["color", "entrance_time"]

    #####################
    # Lifecycle methods #
    #####################

    def setup(self, builder: Builder) -> None:
        self.colors = builder.configuration.population.colors

    ########################
    # Event-driven methods #
    ########################

    def on_initialize_simulants(self, pop_data: SimulantData) -> None:
        new_population = pd.DataFrame(
            {
                "color": np.random.choice(self.colors, len(pop_data.index)),
                "entrance_time": pop_data.creation_time,
            },
            index=pop_data.index,
        )
        self.population_view.update(new_population)



from __future__ import annotations

from typing import Any

import pandas as pd
from layered_config_tree import ConfigurationError

from vivarium import Component, Observer
from vivarium.framework.engine import Builder
from vivarium.framework.event import Event
from vivarium.framework.population import SimulantData
from vivarium.framework.resource import Resource
from vivarium.manager import Manager


class MockComponentA(Observer):
    @property
    def name(self) -> str:
        return self._name

    @property
    def configuration_defaults(self) -> dict[str, Any]:
        return {}

    def __init__(self, *args: Any, name: str = "mock_component_a") -> None:
        super().__init__()
        self._name = name
        self.args = args
        self.builder_used_for_setup = None

    def create_lookup_tables(self, builder: Builder) -> dict[str, Any]:
        return {}

    def register_observations(self, builder: Builder) -> None:
        pass

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self.name == other.name


class MockComponentB(Observer):
    @property
    def name(self) -> str:
        return self._name

    def __init__(self, *args: Any, name: str = "mock_component_b") -> None:
        super().__init__()
        self._name = name
        self.args = args
        self.builder_used_for_setup: Builder | None = None
        if len(self.args) > 1:
            self._sub_components = [MockComponentB(arg, name=arg) for arg in self.args]

    def setup(self, builder: Builder) -> None:
        self.builder_used_for_setup = builder

    def register_observations(self, builder: Builder) -> None:
        builder.results.register_adding_observation(self.name, aggregator=self.counter)

    def create_lookup_tables(self, builder: Builder) -> dict[str, Any]:
        return {}

    def counter(self, _: Any) -> float:
        return 1.0

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self.name == other.name


class MockGenericComponent(Component):
    CONFIGURATION_DEFAULTS = {
        "component": {
            "key1": "val",
            "key2": {
                "subkey1": "val",
                "subkey2": "val",
            },
            "key3": ["val", "val", "val"],
        }
    }

    @property
    def name(self) -> str:
        return self._name

    @property
    def configuration_defaults(self) -> dict[str, Any]:
        return {self.name: self.CONFIGURATION_DEFAULTS["component"]}

    def __init__(self, name: str):
        super().__init__()
        self._name = name
        self.builder_used_for_setup: Builder | None = None

    def setup(self, builder: Builder) -> None:
        self.builder_used_for_setup = builder
        self.config = builder.configuration[self.name]

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self.name == other.name


class Listener(MockComponentB):
    def __init__(self, *args: Any, name: str = "test_listener"):
        super().__init__(*args, name=name)
        self.post_setup_called = False
        self.time_step_prepare_called = False
        self.time_step_called = False
        self.time_step_cleanup_called = False
        self.collect_metrics_called = False
        self.simulation_end_called = False

        self.event_indexes: dict[str, pd.Index[int] | None] = {
            "time_step_prepare": None,
            "time_step": None,
            "time_step_cleanup": None,
            "collect_metrics": None,
        }

    def on_post_setup(self, event: Event) -> None:
        self.post_setup_called = True

    def on_time_step_prepare(self, event: Event) -> None:
        self.time_step_prepare_called = True
        self.event_indexes["time_step_prepare"] = event.index

    def on_time_step(self, event: Event) -> None:
        self.time_step_called = True
        self.event_indexes["time_step"] = event.index

    def on_time_step_cleanup(self, event: Event) -> None:
        self.time_step_cleanup_called = True
        self.event_indexes["time_step_cleanup"] = event.index

    def on_collect_metrics(self, event: Event) -> None:
        self.collect_metrics_called = True
        self.event_indexes["collect_metrics"] = event.index

    def on_simulation_end(self, event: Event) -> None:
        self.simulation_end_called = True


class ColumnCreator(Component):
    @property
    def columns_created(self) -> list[str]:
        return ["test_column_1", "test_column_2", "test_column_3"]

    def setup(self, builder: Builder) -> None:
        builder.value.register_value_producer("pipeline_1", lambda x: x)

    def on_initialize_simulants(self, pop_data: SimulantData) -> None:
        self.population_view.update(self.get_initial_state(pop_data.index))

    def get_initial_state(self, index: pd.Index[int]) -> pd.DataFrame:
        return pd.DataFrame(
            {column: [i % 3 for i in index] for column in self.columns_created}, index=index
        )


class LookupCreator(ColumnCreator):
    CONFIGURATION_DEFAULTS = {
        "lookup_creator": {
            "data_sources": {
                "favorite_team": "simulants.favorite_team",
                "favorite_scalar": 0.4,
                "favorite_color": "simulants.favorite_color",
                "favorite_number": "simulants.favorite_number",
                "baking_time": "self::load_baking_time",
                "cooling_time": "tests.framework.components.test_component::load_cooling_time",
            },
            "alternate_data_sources": {
                "favorite_list": [9, 4],
            },
        }
    }

    def build_all_lookup_tables(self, builder: "Builder") -> None:
        super().build_all_lookup_tables(builder)
        if not self.configuration:
            raise ConfigurationError(
                "Configuration not set. This may break tests using the lookup table creator helper."
            )
        self.lookup_tables["favorite_list"] = self.build_lookup_table(
            builder,
            self.configuration["alternate_data_sources"]["favorite_list"],
            ["column_1", "column_2"],
        )

    @staticmethod
    def load_baking_time(_builder: Builder) -> float:
        return 0.5


class SingleLookupCreator(ColumnCreator):
    pass


class OrderedColumnsLookupCreator(Component):
    @property
    def columns_created(self) -> list[str]:
        return ["foo", "bar"]

    def on_initialize_simulants(self, pop_data: SimulantData) -> None:
        initialization_data = pd.DataFrame(
            {
                "foo": "key1",
                "bar": 15,
            },
            index=pop_data.index,
        )
        self.population_view.update(initialization_data)

    def build_all_lookup_tables(self, builder: "Builder") -> None:
        value_columns = ["one", "two", "three", "four", "five", "six", "seven"]
        ordered_columns = pd.DataFrame(
            [[1, 2, 3, 4, 5, 6, 7], [8, 9, 10, 11, 12, 13, 14]],
            columns=value_columns,
        )
        ordered_columns_categorical = ordered_columns.copy()
        ordered_columns_categorical["foo"] = ["key1", "key2"]
        ordered_columns_interpolated = ordered_columns.copy()
        ordered_columns_interpolated["foo"] = ["key1", "key1"]
        ordered_columns_interpolated["bar_start"] = [10, 20]
        ordered_columns_interpolated["bar_end"] = [20, 30]
        self.lookup_tables["ordered_columns_categorical"] = self.build_lookup_table(
            builder,
            ordered_columns_categorical,
            value_columns,
        )
        self.lookup_tables["ordered_columns_interpolated"] = self.build_lookup_table(
            builder,
            ordered_columns_interpolated,
            value_columns,
        )


class ColumnRequirer(Component):
    @property
    def columns_required(self) -> list[str]:
        return ["test_column_1", "test_column_2"]


class ColumnCreatorAndRequirer(Component):
    @property
    def columns_required(self) -> list[str]:
        return ["test_column_1", "test_column_2"]

    @property
    def columns_created(self) -> list[str]:
        return ["test_column_4"]

    @property
    def initialization_requirements(self) -> list[str | Resource]:
        return ["test_column_2", self.pipeline, self.randomness]

    def setup(self, builder: Builder) -> None:
        self.pipeline = builder.value.get_value("pipeline_1")
        self.randomness = builder.randomness.get_stream("stream_1")

    def on_initialize_simulants(self, pop_data: SimulantData) -> None:
        initialization_data = pd.DataFrame({"test_column_4": 8}, index=pop_data.index)
        self.population_view.update(initialization_data)


class ColumnCreatorAndAllRequirer(ColumnCreatorAndRequirer):
    @property
    def columns_required(self) -> list[str]:
        return []


class AllColumnsRequirer(Component):
    @property
    def columns_required(self) -> list[str]:
        return []


class FilteredPopulationView(ColumnRequirer):
    @property
    def population_view_query(self) -> str:
        return "test_column_1 == 5"


class NoPopulationView(Component):
    pass


class Parameterized(Component):
    def __init__(self, p_one: str, p_two: int, p_three: str):
        super().__init__()
        self.p_one = p_one
        self.p_two = p_two
        self._p_three = p_three


class ParameterizedByComponent(Component):
    def __init__(self, param: Parameterized):
        super().__init__()
        self.param = param


class DefaultPriorities(Component):
    def on_post_setup(self, event: Event) -> None:
        pass

    def on_time_step_prepare(self, event: Event) -> None:
        pass

    def on_time_step(self, event: Event) -> None:
        pass

    def on_time_step_cleanup(self, event: Event) -> None:
        pass

    def on_collect_metrics(self, event: Event) -> None:
        pass

    def on_simulation_end(self, event: Event) -> None:
        pass


class CustomPriorities(DefaultPriorities):
    @property
    def post_setup_priority(self) -> int:
        return 8

    @property
    def time_step_prepare_priority(self) -> int:
        return 7

    @property
    def time_step_priority(self) -> int:
        return 2

    @property
    def time_step_cleanup_priority(self) -> int:
        return 3

    @property
    def collect_metrics_priority(self) -> int:
        return 6

    @property
    def simulation_end_priority(self) -> int:
        return 1


class MockManager(Manager):
    @property
    def name(self) -> str:
        return self._name

    def __init__(self, name: str) -> None:
        self._name = name
